/*
 * Copyright (c) 2006 - 2011 Michael Schierl
 * All rights reserved.
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package net.sf.jmatchparser.template;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.util.List;
import net.sf.jmatchparser.template.engine.template.MatchTemplateImpl;

/**
 * A template, usually loaded from a file, that is used by the {@link Parser}
 * for matching.
 */
public class MatchTemplate {

	private final MatchTemplateImpl impl;

	/**
	 * Load a match template from a file.
	 * 
	 * @param filename
	 *            Name of the file
	 * @param encoding
	 *            File encoding
	 * @param formatters
	 *            Additional formatters to use (optional)
	 */
	public MatchTemplate(String filename, String encoding, Formatter... formatters) throws IOException {
		this(new File(filename), encoding, formatters);
	}

	/**
	 * Load a match template from a file.
	 * 
	 * @param file
	 *            The file
	 * @param encoding
	 *            File encoding
	 * @param formatters
	 *            Additional formatters to use (optional)
	 */
	public MatchTemplate(File file, String encoding, Formatter... formatters) throws IOException {
		this(new FileMatchTemplateResolver(file.getAbsoluteFile().getParentFile(), encoding), file.getName(), null, formatters);
	}

	/**
	 * Load a match template from a resource (a "file" on the class path).
	 * 
	 * @param clazz
	 *            Class to use as a base for finding the resource
	 * @param resourceName
	 *            Name of the resource
	 * @param encoding
	 *            File encoding
	 * @param formatters
	 *            Additional formatters to use (optional)
	 */
	public MatchTemplate(Class<?> clazz, String resourceName, String encoding, Formatter... formatters) throws IOException {
		this(new ResourceMatchTemplateResolver(clazz, resourceName.substring(0, resourceName.lastIndexOf("/") + 1), encoding), resourceName.substring(resourceName.lastIndexOf("/") + 1), null, formatters);
	}

	/**
	 * Load a match template from a custom reader.
	 * 
	 * @param resolver
	 *            Match template resolver for resolving additional template
	 *            names (<code>$INCLUDE</code>), optional
	 * @param templateName
	 *            Name of the template
	 * @param br
	 *            Reader to read the template from
	 * @param formatters
	 *            Additional formatters to use (optional)
	 */
	public MatchTemplate(MatchTemplateResolver resolver, String templateName, BufferedReader br, Formatter... formatters) throws IOException {
		this("plain", resolver, templateName, br, formatters);
	}

	/**
	 * Load a match template from a custom reader using a custom match language.
	 * 
	 * @param matchLanguage
	 *            The match language (currently only <code>"plain"</code> is
	 *            supported)
	 * @param resolver
	 *            Match template resolver for resolving additional template
	 *            names (<tt>$INCLUDE</tt>), optional
	 * @param templateName
	 *            Name of the template
	 * @param br
	 *            Reader to read the template from
	 * @param formatters
	 *            Additional formatters to use (optional)
	 */
	public MatchTemplate(String matchLanguage, MatchTemplateResolver resolver, String templateName, BufferedReader br, Formatter... formatters) throws IOException {
		this.impl = MatchTemplateImpl.createTemplateImpl(matchLanguage, resolver, templateName, br, formatters);
	}

	/**
	 * Enable tracing of covered statements. This can be used to find statements
	 * in your match templates that are no longer needed.
	 * 
	 * @see #getUncoveredStatements()
	 */
	public void enableCoveredStatementsTracing() {
		impl.enableCoveredStatementsTracing();
	}

	/**
	 * Get a list of uncovered statements. This only works if
	 * {@link #enableCoveredStatementsTracing()} has been called before parsing.
	 */
	public List<String> getUncoveredStatements() {
		return impl.getUncoveredStatements();
	}
	
	/**
	 * Get a list of template file names that are used by this template. This
	 * includes the script itself and (recursively) all scripts that have been
	 * included by it, in the order they were included.
	 */
	public List<String> getIncludedTemplateFileNames() {
		return impl.getIncludedTemplateFileNames();
	}

	MatchTemplateImpl getImpl() {
		return impl;
	}

	/**
	 * Interface to resolve additional templates from <code>$INCLUDE</code>
	 * statements.
	 */
	public static interface MatchTemplateResolver {

		/**
		 * Resolve a template.
		 * 
		 * @param templateName
		 *            Name of the template
		 * @return Buffered reader for the template
		 */
		public BufferedReader resolve(String templateName) throws IOException;
	}

	private static class FileMatchTemplateResolver implements MatchTemplateResolver {

		private final File directory;
		private final String encoding;

		public FileMatchTemplateResolver(File directory, String encoding) {
			this.directory = directory;
			this.encoding = encoding;
		}

		@Override
		public BufferedReader resolve(String templateName) throws IOException {
			return new BufferedReader(new InputStreamReader(new FileInputStream(new File(directory, templateName)), encoding));
		}
	}

	private static class ResourceMatchTemplateResolver implements MatchTemplateResolver {

		private final Class<?> clazz;
		private final String resourceBase;
		private final String encoding;

		public ResourceMatchTemplateResolver(Class<?> clazz, String resourceBase, String encoding) {
			this.clazz = clazz;
			this.resourceBase = resourceBase;
			this.encoding = encoding;
		}

		@Override
		public BufferedReader resolve(String templateName) throws IOException {
			return new BufferedReader(new InputStreamReader(clazz.getResourceAsStream(resourceBase + templateName), encoding));
		}
	}
}
