/*
 * Copyright (c) 2006 - 2011 Michael Schierl
 * All rights reserved.
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package net.sf.jmatchparser.template;

import java.io.BufferedReader;
import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.PrintStream;
import java.io.StringWriter;

import javax.xml.bind.JAXBException;
import javax.xml.bind.Unmarshaller;
import javax.xml.bind.annotation.XmlRootElement;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;

import net.sf.jmatchparser.template.engine.ParserImpl;
import net.sf.jmatchparser.util.StreamForwarder;
import net.sf.jmatchparser.util.csv.AbstractCSVWriter;
import net.sf.jmatchparser.util.csv.XMLCSVReader;

import org.w3c.dom.Document;
import org.w3c.dom.Element;

/**
 * A parser for {@link MatchTemplate}s. Each parser can be used for parsing
 * exactly one file/input stream; the templates can be reused for parsing more
 * than one file.
 */
public class Parser {

	private final ParserImpl impl;

	/**
	 * Create a new parser using a new {@link MatchTemplate}.
	 * 
	 * @param filename
	 *            Name of the file
	 * @param encoding
	 *            File encoding
	 * @see MatchTemplate#MatchTemplate(String, String, Formatter...)
	 */
	public Parser(String filename, String encoding) throws IOException {
		this(new MatchTemplate(filename, encoding));
	}

	/**
	 * Create a new parser using a new {@link MatchTemplate}.
	 * 
	 * @param file
	 *            The file
	 * @param encoding
	 *            File encoding
	 * @see MatchTemplate#MatchTemplate(File, String, Formatter...)
	 */
	public Parser(File file, String encoding) throws IOException {
		this(new MatchTemplate(file, encoding));
	}

	/**
	 * Create a new parser using a new {@link MatchTemplate}.
	 * 
	 * @param clazz
	 *            Class to use as a base for finding the resource
	 * @param resourceName
	 *            Name of the resource
	 * @param encoding
	 *            File encoding
	 * @see MatchTemplate#MatchTemplate(Class, String, String, Formatter...)
	 */
	public Parser(Class<?> clazz, String resourceName, String encoding) throws IOException {
		this(new MatchTemplate(clazz, resourceName, encoding));
	}

	/**
	 * Create a new parser using a new {@link MatchTemplate}.
	 * 
	 * @param resolver
	 *            Match template resolver for resolving additional template
	 *            names ($INCLUDE), optional
	 * @param templateName
	 *            Name of the template
	 * @param br
	 *            Reader to read the template from
	 * 
	 * @see MatchTemplate#MatchTemplate(MatchTemplate.MatchTemplateResolver,
	 *      String, BufferedReader, Formatter...)
	 */
	public Parser(MatchTemplate.MatchTemplateResolver resolver, String templateName, BufferedReader br) throws IOException {
		this(new MatchTemplate(resolver, templateName, br));
	}

	/**
	 * Create a new parser using an existing {@link MatchTemplate}.
	 * 
	 * @param template
	 *            The match template to use
	 */
	public Parser(MatchTemplate template) {
		this(new ParserImpl(template.getImpl()));
	}

	private Parser(ParserImpl impl) {
		this.impl = impl;
	}

	/**
	 * Set a local variable in this parser. This method can only be called
	 * before parsing started.
	 * 
	 * @param name
	 *            Name of the variable
	 * @param value
	 *            Value of the variable
	 */
	public void setLocal(String name, String value) {
		impl.setLocal(name, value);
	}

	/**
	 * Set the name of this parser. This name is used in error messages and can
	 * be used to distinguish multiple parsers easily.
	 */
	public void setParserName(String parserName) {
		impl.setParserName(parserName);
	}

	/**
	 * Parse an input stream using a JAXB {@link Unmarshaller} for parsing the
	 * resulting XML.
	 * 
	 * @param <T>
	 *            Type of the unmarshalled object
	 * @param stream
	 *            Input stream to parse
	 * @param encoding
	 *            Encoding of the stream
	 * @param unmarshaller
	 *            {@link Unmarshaller} to use
	 * @param clazz
	 *            Class of the unmarshalled object
	 * @return Unmarshalled object
	 */
	public <T> T parse(InputStream stream, String encoding, Unmarshaller unmarshaller, Class<T> clazz) throws ParserException, ParserConfigurationException, IOException, JAXBException {
		String name = clazz.getAnnotation(XmlRootElement.class).name();
		Document doc = parse(stream, encoding, name);
		return unmarshaller.unmarshal(doc.getDocumentElement(), clazz).getValue();
	}

	/**
	 * Parse an input stream and create a new DOM {@link Document} from the
	 * resulting XML.
	 * 
	 * @param stream
	 *            Input stream to parse
	 * @param encoding
	 *            Encoding of the stream
	 * @param toplevelTag
	 *            Tag of the document element to use in the new document
	 * @return {@link Document}
	 */
	public Document parse(InputStream stream, String encoding, String toplevelTag) throws ParserException, ParserConfigurationException, IOException {
		BufferedReader br = new BufferedReader(new InputStreamReader(stream, encoding));
		return parse(br, toplevelTag);
	}

	/**
	 * Parse an input stream and add the resulting XML to a DOM {@link Element}.
	 * 
	 * @param stream
	 *            Input stream to parse
	 * @param encoding
	 *            Encoding of the stream
	 * @param target
	 *            Element to add the XML to
	 */
	public void parse(InputStream stream, String encoding, Element target) throws ParserException, IOException {
		BufferedReader br = new BufferedReader(new InputStreamReader(stream, encoding));
		parse(br, target);
	}

	/**
	 * Parse a buffered reader using a JAXB {@link Unmarshaller} for parsing the
	 * resulting XML.
	 * 
	 * @param <T>
	 *            Type of the unmarshalled object
	 * @param reader
	 *            {@link BufferedReader} to parse
	 * @param unmarshaller
	 *            {@link Unmarshaller} to use
	 * @param clazz
	 *            Class of the unmarshalled object
	 * @return Unmarshalled object
	 */
	public <T> T parse(BufferedReader reader, Unmarshaller unmarshaller, Class<T> clazz) throws ParserException, ParserConfigurationException, IOException, JAXBException {
		String name = clazz.getAnnotation(XmlRootElement.class).name();
		Document doc = parse(reader, name);
		return unmarshaller.unmarshal(doc.getDocumentElement(), clazz).getValue();
	}

	/**
	 * Parse a buffered reader and create a new DOM {@link Document} from the
	 * resulting XML.
	 * 
	 * @param reader
	 *            {@link BufferedReader} to parse
	 * @param toplevelTag
	 *            Tag of the document element to use in the new document
	 * @return {@link Document}
	 */
	public Document parse(BufferedReader reader, String toplevelTag) throws ParserException, ParserConfigurationException, IOException {
		Document doc = DocumentBuilderFactory.newInstance().newDocumentBuilder().newDocument();
		doc.appendChild(doc.createElement(toplevelTag));
		parse(reader, doc.getDocumentElement());
		return doc;
	}

	/**
	 * Parse a buffered reader and add the resulting XML to a DOM
	 * {@link Element}.
	 * 
	 * @param reader
	 *            {@link BufferedReader} to parse
	 * @param target
	 *            Element to add the XML to
	 */
	public void parse(BufferedReader reader, Element target) throws ParserException, IOException {
		StringWriter sw = new StringWriter();
		StreamForwarder.forward(reader, sw);
		String file = sw.toString().replace("\r\n", "\n").replace('\r', '\n');
		if (!file.endsWith("\n"))
			file += "\n";
		parse(file, target);
	}

	/**
	 * Parse a {@link String} or other char sequence using a JAXB
	 * {@link Unmarshaller} for parsing the resulting XML.
	 * 
	 * @param <T>
	 *            Type of the unmarshalled object
	 * @param file
	 *            {@link CharSequence} to parse
	 * @param unmarshaller
	 *            {@link Unmarshaller} to use
	 * @param clazz
	 *            Class of the unmarshalled object
	 * @return Unmarshalled object
	 */
	public <T> T parse(CharSequence file, Unmarshaller unmarshaller, Class<T> clazz) throws ParserException, ParserConfigurationException, IOException, JAXBException {
		String name = clazz.getAnnotation(XmlRootElement.class).name();
		Document doc = parse(file, name);
		return unmarshaller.unmarshal(doc.getDocumentElement(), clazz).getValue();
	}

	/**
	 * Parse a {@link String} or other char sequence and create a new DOM
	 * {@link Document} from the resulting XML.
	 * 
	 * @param file
	 *            {@link CharSequence} to parse
	 * @param toplevelTag
	 *            Tag of the document element to use in the new document
	 * @return {@link Document}
	 */
	public Document parse(CharSequence file, String toplevelTag) throws ParserException, ParserConfigurationException, IOException {
		Document doc = DocumentBuilderFactory.newInstance().newDocumentBuilder().newDocument();
		doc.appendChild(doc.createElement(toplevelTag));
		parse(file, doc.getDocumentElement());
		return doc;
	}

	/**
	 * Parse a {@link String} or other char sequence and add the resulting XML
	 * to a DOM {@link Element}.
	 * 
	 * @param file
	 *            {@link CharSequence} to parse
	 * @param target
	 *            Element to add the XML to
	 */
	public void parse(CharSequence file, Element target) throws ParserException, IOException {
		impl.parse(file, target);
	}

	/**
	 * Parse an input stream to a CSV file.
	 * 
	 * @param stream
	 *            Input stream to parse
	 * @param encoding
	 *            Encoding of the stream
	 * @param writer
	 *            {@link AbstractCSVWriter} to write the file to
	 */
	public void parseToCSV(InputStream stream, String encoding, AbstractCSVWriter writer) throws ParserConfigurationException, ParserException, IOException {
		convertDOMToCSV(parse(stream, encoding, "csv"), writer);
	}

	/**
	 * Parse a buffered reader to a CSV file.
	 * 
	 * @param reader
	 *            {@link BufferedReader} to parse
	 * @param writer
	 *            {@link AbstractCSVWriter} to write the file to
	 */
	public void parseToCSV(BufferedReader reader, AbstractCSVWriter writer) throws ParserConfigurationException, ParserException, IOException {
		convertDOMToCSV(parse(reader, "csv"), writer);
	}

	/**
	 * Parse a {@link String} or other char sequence to a CSV file.
	 * 
	 * @param file
	 *            {@link CharSequence} to parse
	 * @param writer
	 *            {@link AbstractCSVWriter} to write the file to
	 */

	public void parseToCSV(CharSequence file, AbstractCSVWriter writer) throws ParserConfigurationException, ParserException, IOException {
		convertDOMToCSV(parse(file, "csv"), writer);
	}

	private static void convertDOMToCSV(Document doc, AbstractCSVWriter writer) throws IOException {
		convertDOMToCSV(doc, new AbstractCSVWriter[] { writer });
	}

	/**
	 * Convert a DOM {@link Document} to one or more CSV files. Each child
	 * element of the document element is a row, each child element of a row is
	 * a cell. Rows may have an attribute <code>@file</code> which contains the
	 * zero-based index of the writer to save this row to (zero by default); all
	 * other attributes are ignored.
	 * 
	 * @param doc
	 *            {@link Document} to parse
	 * @param writers
	 *            {@link AbstractCSVWriter}(s) to write the file(s) to
	 */
	public static void convertDOMToCSV(Document doc, AbstractCSVWriter[] writers) throws IOException {
		XMLCSVReader r = new XMLCSVReader(doc);
		String[] record;
		int[] fileIndex = new int[1];
		while ((record = r.read(fileIndex)) != null) {
			writers[fileIndex[0]].write(record);
			fileIndex[0] = 0;
		}
	}

	/**
	 * Add a callback function to this parser.
	 * 
	 * @param name
	 *            Name of the function
	 * @param function
	 *            Function to add
	 */
	public void addCallback(String name, CallbackFunction function) {
		impl.addCallback(name, function);
	}

	/**
	 * Set the debug stream where <code>DEBUGWRITE</code> and
	 * <code>DUMPLINE</code> commands write to.
	 * 
	 * @param debugStream
	 *            New debug stream to use
	 */
	public void setDebugStream(PrintStream debugStream) {
		impl.setDebugStream(debugStream);
	}
}
