/*
 * Copyright (c) 2006 - 2011 Michael Schierl
 * All rights reserved.
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package net.sf.jmatchparser.template.engine;

import java.io.IOException;
import java.io.PrintStream;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import net.sf.jmatchparser.template.CallbackFunction;
import net.sf.jmatchparser.template.ParserException;
import net.sf.jmatchparser.template.engine.operation.Operation;
import net.sf.jmatchparser.template.engine.operation.Operation.OperationResult;

import net.sf.jmatchparser.template.engine.template.MatchTemplateImpl;
import org.w3c.dom.Element;

public class ParserImpl {

	private CharSequence file = null;
	private StringBuilder writeBuffer = new StringBuilder();
	private String parserName = null;

	private List<ParserState> queuedStates = new ArrayList<ParserState>();
	private final MatchTemplateImpl template;
	private Map<String, CallbackFunction> callbacks = new HashMap<String, CallbackFunction>();
	private ParserState sideEffectState = null;
	private PrintStream debugStream = System.out;

	public ParserImpl(MatchTemplateImpl template) {
		this.template = template;
		queuedStates.add(new ParserState(this));
		DefaultCallbackFunctions.addTo(this);
	}

	public void setLocal(String name, String value) {
		if (file != null)
			throw new IllegalStateException("Parser has already been used");
		queuedStates.get(0).setLocal(name, value);
	}

	public void setParserName(String parserName) {
		this.parserName = parserName;
	}

	public void parse(CharSequence file, Element target) throws ParserException, IOException {
		if (this.file != null)
			throw new IllegalStateException("Parser has already been used");
		this.file = file;
		ParserState finalState = run();
		finalState.applyXMLActions(target);
		if (finalState.getCallStack() != null)
			throw new IllegalStateException("Pending return calls");
		if (finalState.getCoveredOperations() != null)
			template.addCoveredOperations(finalState.getCoveredOperations());
	}

	private ParserState run() throws ParserException, IOException {
		int lastFailOffset = -1;
		List<String> lastFailTemplatePositions = new ArrayList<String>();
		List<String> sideEffectBlockedTemplatePositions = new ArrayList<String>();
		ParserState finishedState = null;
		while (queuedStates.size() > 0) {
			ParserState current = queuedStates.remove(0);
			OperationResult opres = OperationResult.CONTINUE;
			while (opres == OperationResult.CONTINUE) {
				int ip = current.getNextInstructionAndIncrement();
				Operation op = template.getOperation(ip);
				try {
					opres = op.execute(current);
				} catch (RuntimeException ex) {
					throw new RuntimeException("Cannot execute operation at template position:\n\t" + op.getTemplatePosition(), ex);
				}
				if (opres == null)
					throw new RuntimeException(op.getClass().toString() + " returned invalid operation result");
				switch (opres) {
				case CONTINUE:
					// everything fine
					break;
				case NEXTPASS:
					lastFailOffset = -1;
					lastFailTemplatePositions.clear();
					sideEffectBlockedTemplatePositions.clear();
					opres = OperationResult.CONTINUE;
					break;
				case DIE:
					if (queuedStates.size() == 0)
						throw new RuntimeException("Last state died");
					break;
				case FAIL_BLOCKED:
					sideEffectBlockedTemplatePositions.add(op.getTemplatePosition());
					// fall through
				case FAIL:
					if (current.getOffset() > lastFailOffset) {
						lastFailOffset = current.getOffset();
						lastFailTemplatePositions.clear();
					}
					if (current.getOffset() == lastFailOffset) {
						String cmd = op.getTemplatePosition();
						if (!lastFailTemplatePositions.contains(cmd)) {
							lastFailTemplatePositions.add(cmd);
						}
					}
					break;
				case FINISHED:
					if (finishedState != null) {
						throw new IOException("Multiple paths finished successfully.");
					}
					finishedState = current;
					break;
				}
			}
		}
		if (finishedState == null) {
			if (lastFailOffset == -1)
				throw new RuntimeException("No fail offset recorded!");
			StringBuilder sb = new StringBuilder();
			sb.append("Could not parse file" + (getParserName() == null ? "" : " '" + getParserName() + "'") + ". Template position(s):");
			for (String templatePosition : lastFailTemplatePositions) {
				sb.append("\n\t" + templatePosition);
			}
			if (sideEffectBlockedTemplatePositions.size() > 0) {
				sb.append("\nSide effects blocked:");
				for (String templatePosition : sideEffectBlockedTemplatePositions) {
					sb.append("\n\t" + templatePosition);
				}
			}
			sb.append("\nNext 1K to parse:\n\t");
			sb.append(getFile().subSequence(lastFailOffset, Math.min(getFile().length(), lastFailOffset + 1024)).toString().replace("\n", "\n\t"));
			throw new ParserException(sb.toString());
		}
		return finishedState;
	}

	public CharSequence getFile() {
		return file;
	}

	public String getBuffer() {
		String result = writeBuffer.toString();
		writeBuffer.setLength(0);
		return result;
	}

	public void writeBuffer(String text) {
		if (hasQueuedStates())
			throw new RuntimeException("There are queued states!");
		writeBuffer.append(text);
	}

	public MatchTemplateImpl getTemplate() {
		return template;
	}

	public boolean hasQueuedStates() {
		return queuedStates.size() > 0;
	}

	public void enqueueState(ParserState toEnqueue) {
		queuedStates.add(toEnqueue);
	}

	protected void setFile(ParserState current, CharSequence newFile) {
		if (hasQueuedStates())
			throw new RuntimeException("There are queued states!");
		if (current.getOffset() != 0)
			throw new RuntimeException("Some text has already been parsed:" + current.getOffset());
		file = newFile;
	}

	public void addCallback(String name, CallbackFunction function) {
		callbacks.put(name, function);
	}

	public CallbackFunction getCallback(String name) {
		return callbacks.get(name);
	}

	public String getParserName() {
		return parserName;
	}

	public ParserState getSideEffectState() {
		return sideEffectState;
	}

	public void setSideEffectState(ParserState sideEffectState) {
		this.sideEffectState = sideEffectState;
	}

	public PrintStream getDebugStream() {
		return debugStream;
	}

	public void setDebugStream(PrintStream debugStream) {
		this.debugStream = debugStream;
	}
}
