/*
 * Copyright (c) 2006 - 2011 Michael Schierl
 * All rights reserved.
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package net.sf.jmatchparser.template.engine;

import java.util.ArrayList;
import java.util.BitSet;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.w3c.dom.Element;

public class ParserState {
	private ParserImpl parser;
	private int nextInstruction;
	private int offset;
	private Map<String, String> locals = new HashMap<String, String>();
	private List<String[]> xmlActions = new ArrayList<String[]>();
	private final List<String[]> preprocessedXmlActions;
	private final BitSet coveredOperations;
	private CallStackFrame callStack;

	public ParserState(ParserImpl parser) {
		this.parser = parser;
		nextInstruction = 0;
		offset = 0;
		preprocessedXmlActions = new ArrayList<String[]>();
		if (parser.getTemplate().getCoveredStatementsTracingSize() != 0) {
			coveredOperations = new BitSet(parser.getTemplate().getCoveredStatementsTracingSize());
		} else {
			coveredOperations = null;
		}
	}

	public ParserState(ParserState toCopy, int nextInstruction) {
		xmlActions = new ArrayList<String[]>(toCopy.xmlActions);
		preprocessedXmlActions = toCopy.preprocessedXmlActions; // we can share
																// them
		parser = toCopy.parser;
		this.nextInstruction = nextInstruction;
		offset = toCopy.offset;
		locals.putAll(toCopy.locals);
		callStack = toCopy.callStack;
		if (toCopy.coveredOperations == null) {
			coveredOperations = null;
		} else {
			coveredOperations = (BitSet) toCopy.coveredOperations.clone();
		}
	}

	public void applyXMLActions(Element target) {
		checkNoQueuedStates();
		if (parser != null && parser.getBuffer().length() > 0)
			throw new IllegalStateException();
		Element current = target;
		preprocessXmlActions();
		for (String[] elem : preprocessedXmlActions) {
			switch (elem[0].charAt(0)) {
			case '/':
				current = (Element) current.getParentNode();
				break;
			case '<':
				Element e = current.getOwnerDocument().createElement(elem[1]);
				current.appendChild(e);
				current = e;
				break;
			case 'T':
				current.appendChild(current.getOwnerDocument().createTextNode(elem[1]));
				break;
			case '@':
				current.setAttribute(elem[1], elem[2]);
				break;
			default:
				throw new RuntimeException();
			}
		}
		if (current != target)
			throw new IllegalStateException("Incorrect nesting of XML actions");
		xmlActions.clear();
	}

	public void appendXmlAction(String... action) {
		xmlActions.add(action);
	}

	public String getLocal(String varname) {
		return locals.get(varname);
	}

	public Map<String, String> getLocals() {
		return new HashMap<String, String>(locals);
	}

	public void setLocal(String varname, String value) {
		locals.put(varname, value);
		if (value == null)
			locals.remove(varname);
	}

	public int getNextInstructionAndIncrement() {
		if (coveredOperations != null)
			coveredOperations.set(nextInstruction);
		return nextInstruction++;
	}

	public int getNextInstruction() {
		return nextInstruction;
	}

	public void setNextInstruction(int nextInstruction) {
		this.nextInstruction = nextInstruction;
	}

	public int getOffset() {
		return offset;
	}

	public ParserImpl getParser() {
		return parser;
	}

	public CharSequence getRest() {
		return parser.getFile().subSequence(offset, parser.getFile().length());
	}

	public CharSequence getFirstLine() {
		CharSequence rest = getRest();
		for (int i = 0; i < rest.length(); i++) {
			if (rest.charAt(i) == '\n') {
				return rest.subSequence(0, i);
			}
		}
		return null;
	}

	public void discardCharacters(int count) {
		offset += count;
	}

	public void write(String text) {
		parser.writeBuffer(text);
	}

	public void startNextPass() {
		checkNoQueuedStates();
		if (xmlActions.size() != 0)
			throw new IllegalStateException();
		if (preprocessedXmlActions.size() != 0)
			throw new IllegalStateException();
		if (offset != parser.getFile().length())
			throw new IllegalStateException();
		offset = 0;
		parser.setFile(this, parser.getBuffer());
	}

	private void checkNoQueuedStates() {
		if (parser != null && parser.hasQueuedStates())
			throw new IllegalStateException();
	}

	public void preprocessXmlActions() {
		checkNoQueuedStates();
		preprocessedXmlActions.addAll(xmlActions);
		xmlActions.clear();
	}

	public BitSet getCoveredOperations() {
		return coveredOperations;
	}

	public CallStackFrame getCallStack() {
		return callStack;
	}

	public void setCallStack(CallStackFrame callStack) {
		this.callStack = callStack;
	}
}
