/*
 * Copyright (c) 2006 - 2011 Michael Schierl
 * All rights reserved.
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package net.sf.jmatchparser.template.engine.formatter;

import java.io.IOException;
import java.io.PrintStream;
import java.io.StringReader;
import java.util.Arrays;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;

import net.sf.jmatchparser.template.Formatter;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.xml.sax.EntityResolver;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;

public class XMLFormatter implements Formatter {

	@Override
	public String getName() {
		return "xml";
	}

	@Override
	public String format(String content, PrintStream debugStream) throws IOException {
		try {
			DocumentBuilder dbf = DocumentBuilderFactory.newInstance().newDocumentBuilder();
			dbf.setEntityResolver(new EntityResolver() {
				@Override
				public InputSource resolveEntity(String publicId, String systemId) throws SAXException, IOException {
					if (systemId.startsWith("http://www.w3.org/"))
						return new InputSource(new StringReader(""));
					return null;
				}
			});
			Document doc = dbf.parse(new InputSource(new StringReader(content)));
			StringBuilder result = new StringBuilder();
			write(doc.getDocumentElement(), result);
			return result.toString();
		} catch (ParserConfigurationException ex) {
			throw new IOException(ex);
		} catch (SAXException ex) {
			throw new IOException("SAXException while parsing content:\n" + content, ex);
		}
	}

	private void write(Element element, StringBuilder result) {
		// append start tag
		result.append("<" + quote(element.getTagName(), false));
		if (element.getNamespaceURI() != null)
			result.append(" xmlns=\"" + quote(element.getNamespaceURI(), true));
		String[] attrs = new String[element.getAttributes().getLength()];
		for (int i = 0; i < attrs.length; i++) {
			attrs[i] = element.getAttributes().item(i).getNodeName();
			String ns = element.getAttributes().item(i).getNamespaceURI();
			if (ns != null)
				attrs[i] += "\0" + ns;
		}
		Arrays.sort(attrs);
		for (int i = 0; i < attrs.length; i++) {
			String localName = attrs[i];
			String ns = null;
			if (localName.contains("\0")) {
				ns = localName.substring(localName.indexOf('\0') + 1);
				localName = localName.substring(0, localName.indexOf('\0'));
			}
			if (ns == null) {
				result.append(" " + quote(localName, false));
			} else {
				result.append(" xmlns:attr" + i + "=\"" + quote(ns, true) + "\" attr" + i + ":" + quote(localName, true));
			}
			result.append("=\"" + quote(element.getAttributeNodeNS(ns, localName).getValue(), true) + "\"");
		}
		result.append(">\n");

		// append content
		NodeList children = element.getChildNodes();
		for (int i = 0; i < children.getLength(); i++) {
			Node node = children.item(i);
			switch (node.getNodeType()) {

			case Node.ELEMENT_NODE:
				write((Element) node, result);
				break;

			case Node.TEXT_NODE:
				result.append(quoteMultiline(node, "-"));
				break;

			case Node.CDATA_SECTION_NODE:
				result.append(quoteMultiline(node, "["));
				break;

			case Node.COMMENT_NODE:
				result.append(quoteMultiline(node, "!"));
				break;

			case Node.PROCESSING_INSTRUCTION_NODE:
				result.append(quoteMultiline(node, "?" + node.getNodeName() + " "));
				break;

			case Node.ATTRIBUTE_NODE:
				throw new RuntimeException("Attribute nodes should never be children");

			case Node.ENTITY_REFERENCE_NODE:
			case Node.ENTITY_NODE:
			case Node.DOCUMENT_NODE:
			case Node.DOCUMENT_TYPE_NODE:
			case Node.DOCUMENT_FRAGMENT_NODE:
			case Node.NOTATION_NODE:
				throw new RuntimeException("Unparsable node found: " + node.getNodeType());
			}
		}

		// append end tag
		result.append("</" + quote(element.getTagName(), false) + ">\n");
	}

	private static String quoteMultiline(Node node, String marker) {
		String value = "\n+" + node.getNodeValue().replaceAll("\n", "\n+");
		int last = value.lastIndexOf("\n+");
		value = value.substring(0, last + 1) + marker + value.substring(last + 2);
		String str = quote(value.substring(1), false) + "\n";
		return str;
	}

	private static String quote(String name, boolean quot) {
		if (name.matches("^[^<>&\"]+$"))
			return name;
		StringBuilder result = new StringBuilder();
		for (int i = 0; i < name.length(); i++) {
			switch (name.charAt(i)) {
			case '&':
				result.append("&amp;");
				break;
			case '"':
				result.append(quot ? "&quot;" : "\"");
				break;
			case '<':
				result.append("&lt;");
				break;
			case '>':
				result.append("&gt;");
				break;
			default:
				result.append(name.charAt(i));
				break;
			}
		}
		return result.toString();
	}

}
