/*
 * Copyright (c) 2006 - 2011 Michael Schierl
 * All rights reserved.
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package net.sf.jmatchparser.template.engine.parameter;

import java.util.ArrayList;
import java.util.List;
import java.util.regex.Matcher;

import net.sf.jmatchparser.template.engine.ParserState;

public class Parameter {

	public static Parameter parseParameter(String text, String[] specialtags) {
		return parseParameters(text, 1, 1, 0, specialtags)[0];
	}

	public static Parameter parseRawParameter(String rawText) {
		return parseRawParameters(rawText, 1, 1, 0)[0];
	}

	public static Parameter[] parseParameters(String text, int minparamcount, int maxparamcount, int noCommaCount, String[] specialtags) {
		if (text.indexOf('\1') != -1 || text.indexOf('\2') != -1)
			throw new RuntimeException("Invalid control characters found");
		String start = specialtags[0], end = specialtags[1];
		String startentity = specialtags[2], endentity = specialtags[3];
		if (startentity == null)
			startentity = "\1[\2";
		if (endentity == null)
			endentity = "\1]\2";
		text = text.replace(start, "\1").replace(end, "\2");
		text = text.replace(startentity, start).replace(endentity, end);
		return parseRawParameters(text, minparamcount, maxparamcount, noCommaCount);
	}

	public static Parameter[] parseRawParameters(String rawText, int minparamcount, int maxparamcount, int noCommaCount) {
		if (rawText.indexOf('\0') != -1)
			throw new RuntimeException("Invalid control characters found");
		Parameter[] result = new Parameter[maxparamcount];
		int current = 0;
		StringBuilder literalBuffer = new StringBuilder();
		List<Tag> elems = new ArrayList<Tag>();
		while (true) {
			int pos;
			if (noCommaCount > 0) {
				pos = literalBuffer.indexOf(" ");
				int pos2 = literalBuffer.indexOf("\t");
				if (pos2 != -1 && (pos == -1 || pos > pos2))
					pos = pos2;
				if (pos != -1) {
					literalBuffer.setCharAt(pos, '\0');
				}
			}
			pos = literalBuffer.indexOf("\0");
			if (pos != -1 && current < maxparamcount - 1) {
				if (pos != 0)
					elems.add(new LiteralTag(literalBuffer.substring(0, pos)));
				literalBuffer.delete(0, pos + 1);
				result[current++] = new Parameter(elems);
				elems.clear();
				if (noCommaCount > 0)
					noCommaCount--;
				continue; // check for another parameter
			}
			if (rawText.length() == 0)
				break;
			pos = rawText.indexOf("\1");
			if (pos == -1) {
				literalBuffer.append(rawText);
				rawText = "";
				continue;
			} else if (pos > 0) {
				literalBuffer.append(rawText.substring(0, pos));
				rawText = rawText.substring(pos);
				continue;
			} else {
				rawText = rawText.substring(1);
			}
			int depth = 1;
			for (int i = 0; i < rawText.length(); i++) {
				if (rawText.charAt(i) == '\1')
					depth++;
				if (rawText.charAt(i) == '\2')
					depth--;
				if (depth == 0) {
					Tag pe = Tag.parseTag(rawText.substring(0, i));
					if (pe instanceof LiteralTag) {
						literalBuffer.append(((LiteralTag) pe).getString());
					} else {
						if (literalBuffer.length() > 0) {
							elems.add(new LiteralTag(literalBuffer.substring(0)));
							literalBuffer.delete(0, literalBuffer.length());
						}
						elems.add(pe);
					}
					rawText = rawText.substring(i + 1);
					break;
				}
			}
			if (depth != 0)
				throw new RuntimeException("Unterminated tag.");
		}
		if (literalBuffer.length() > 0)
			elems.add(new LiteralTag(literalBuffer.toString()));
		if (elems.size() > 0 || current != 0 || minparamcount == 1)
			result[current++] = new Parameter(elems);
		if (current < minparamcount)
			throw new RuntimeException("Invalid number of parameters!");
		return result;
	}

	private final Tag[] tags;
	private final String cachedString;
	private final String cachedRegex;

	private Parameter(List<Tag> elems) {
		this(elems.toArray(new Tag[elems.size()]));
	}

	private Parameter(Tag[] elements) {
		this.tags = elements;
		StringBuilder sbRegex = new StringBuilder();
		StringBuilder sbString = new StringBuilder();
		for (Tag e : elements) {
			if (e.isStatic()) {
				sbRegex.append(e.getRegex(null));
				if (e.isParsingOnly() || sbString == null)
					sbString = null;
				else
					sbString.append(e.getString(null));
			} else {
				sbString = sbRegex = null;
				break;
			}
		}
		cachedRegex = (sbRegex == null) ? null : sbRegex.toString();
		cachedString = (sbString == null) ? null : sbString.toString();
	}

	public String getString(ParserState state) {
		if (cachedString != null)
			return cachedString;
		StringBuilder sb = new StringBuilder();
		for (Tag e : tags) {
			sb.append(e.getString(state));
		}
		return sb.toString();
	}

	public String getRegex(ParserState state) {
		if (cachedRegex != null)
			return cachedRegex;
		StringBuilder sb = new StringBuilder();
		for (Tag e : tags) {
			sb.append(e.getRegex(state));
		}
		return sb.toString();
	}

	public boolean isStringStatic() {
		return cachedString != null;
	}

	public void applyMatch(ParserState state, Matcher matcher) {
		int idx = 0;
		for (Tag e : tags) {
			int cnt = e.getUsedGroupCount();
			String[] params = new String[cnt];
			for (int i = 0; i < params.length; i++) {
				params[i] = matcher.group(i + idx + 1);
			}
			e.applyMatch(state, params);
			idx += cnt;
		}
		if (idx != matcher.groupCount())
			throw new RuntimeException("Wrong number of regex groups");
	}

	public static Parameter getLiteralParameter(String literal) {
		return new Parameter(new Tag[] { new LiteralTag(literal) });
	}
}
