/*
 * Copyright (c) 2006 - 2011 Michael Schierl
 * All rights reserved.
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package net.sf.jmatchparser.template.engine.parameter;

import java.util.regex.Matcher;
import java.util.regex.Pattern;

import net.sf.jmatchparser.template.engine.ParserState;
import net.sf.jmatchparser.util.PatternCache;

abstract class Tag {

	boolean isParsingOnly() {
		return false;
	}

	int getUsedGroupCount() {
		return 0;
	}

	String getRegex(ParserState state) {
		return Pattern.quote(getString(state));
	}

	void applyMatch(ParserState state, String[] matches) {
		if (matches.length != 0)
			throw new RuntimeException();
	}

	/**
	 * Return whether both the {@link #getString(ParserState)} and
	 * {@link #getRegex(ParserState)} method will always return the same value,
	 * independent from the {@link ParserState}.
	 */
	abstract boolean isStatic();

	abstract String getString(ParserState state);

	static Tag parseTag(String rawContent) {
		switch (rawContent.length() > 0 ? rawContent.charAt(0) : ' ') {
		case '&':
			return new EntityWrapperTag(parseTag(rawContent.substring(1)));
		case '=':
			return new InsertLocalTag(Parameter.parseRawParameter(rawContent.substring(1)));
		case '^':
			return new CallbackFunctionTag(Parameter.parseRawParameters(rawContent.substring(1), 0, 255, 0));
		case '\\':
			char escaped;
			if (rawContent.length() == 2) {
				switch (rawContent.charAt(1)) {
				case 'r':
					escaped = '\r';
					break;
				case 'n':
					escaped = '\n';
					break;
				case 'b':
					escaped = '\b';
					break;
				case 't':
					escaped = '\t';
					break;
				case '[':
					escaped = '';
					break;
				case ']':
					escaped = '';
					break;
				case '\\':
					escaped = '\\';
					break;
				case '(':
					return new LiteralTag("<<");
				case ')':
					return new LiteralTag(">>");
				default:
					escaped = (char) Integer.parseInt(rawContent.substring(1), 16);
					break;
				}
			} else {
				escaped = (char) Integer.parseInt(rawContent.substring(1), 16);
			}
			return new LiteralTag("" + escaped);
		case ',':
			if (rawContent.length() > 1)
				throw new RuntimeException("Comma tag may not have any parameters");
			return new LiteralTag("\0");
		case '?':
			return new OptionalTag(Parameter.parseRawParameter(rawContent.substring(1)));
		default:
			// case '/':
			// case '@':
			Matcher m = PatternCache.compile("[/@+]?((?:[a-zA-Z][a-zA-Z0-9]*)?)(:.*)?").matcher(rawContent);
			if (!m.matches())
				throw new RuntimeException("Unsupported tag: " + rawContent);
			String name = m.group(1);
			String regex = m.group(2);
			if (regex == null)
				regex = ".*?";
			else
				regex = regex.substring(1);
			if (rawContent.startsWith("/")) {
				return new TagParseTag(name, regex);
			} else if (rawContent.startsWith("@")) {
				return new AttributeParseTag(name, regex);
			} else if (rawContent.startsWith("+")) {
				return new LocalParseTag(name, regex, true);
			} else {
				return new LocalParseTag(name, regex, false);
			}
		}
	}
}
