/*
 * Copyright (c) 2006 - 2011 Michael Schierl
 * All rights reserved.
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package net.sf.jmatchparser.template.engine.template;

import java.io.BufferedReader;
import java.io.IOException;
import java.util.ArrayList;
import java.util.BitSet;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import net.sf.jmatchparser.template.ChainedFormatter;
import net.sf.jmatchparser.template.Formatter;
import net.sf.jmatchparser.template.MatchTemplate.MatchTemplateResolver;
import net.sf.jmatchparser.template.engine.formatter.CSVFormatter;
import net.sf.jmatchparser.template.engine.formatter.JTidyFormatter;
import net.sf.jmatchparser.template.engine.formatter.TrimFormatter;
import net.sf.jmatchparser.template.engine.formatter.XMLFormatter;
import net.sf.jmatchparser.template.engine.operation.Label;
import net.sf.jmatchparser.template.engine.operation.NopOperation;
import net.sf.jmatchparser.template.engine.operation.Operation;

public abstract class MatchTemplateImpl {

	private String currentTemplatePosition;
	private final List<Operation> operations = new ArrayList<Operation>();
	private BitSet coveredOperations = null;
	private final Map<String, Formatter> formatters = new HashMap<String, Formatter>();
	private final Map<String, Label> defLabels = new HashMap<String, Label>();
	protected final List<String> includedTemplateFileNames = new ArrayList<String>();

	protected MatchTemplateImpl(Formatter[] formatters) throws IOException {
		addFormatters(new Formatter[] {
				new TrimFormatter(false),
				new TrimFormatter(true),
				new XMLFormatter(),
				new JTidyFormatter(false),
				new JTidyFormatter(true),
				new CSVFormatter() });
		addFormatters(new Formatter[] { new ChainedFormatter("html", getFormatter("jtidy+cleanspace"), getFormatter("xml")) });
		addFormatters(formatters);
	}

	private void addFormatters(Formatter[] fs) {
		for (Formatter f : fs) {
			formatters.put(f.getName(), f);
		}
	}

	public Formatter getFormatter(String name) {
		return formatters.get(name.toLowerCase().trim());
	}

	public int getNextOperationIndex() {
		return operations.size();
	}

	public void appendOperation(Operation op) {
		operations.add(op);
	}

	public Operation getOperation(int ip) {
		return operations.get(ip);
	}

	public void enableCoveredStatementsTracing() {
		coveredOperations = new BitSet(operations.size());
	}

	public int getCoveredStatementsTracingSize() {
		return coveredOperations == null ? 0 : operations.size();
	}

	public void addCoveredOperations(BitSet operations) {
		coveredOperations.or(operations);
	}

	public List<String> getUncoveredStatements() {
		if (coveredOperations == null)
			throw new IllegalStateException("Covered statements tracing is not enabled");
		List<String> result = new ArrayList<String>();
		for (int i = 0; i < operations.size(); i++) {
			if (!coveredOperations.get(i)) {
				String tpos = operations.get(i).getTemplatePosition();
				if (!result.contains(tpos))
					result.add(tpos);
			}
		}
		return result;
	}
	
	public List<String> getIncludedTemplateFileNames() {
		return includedTemplateFileNames;
	}

	public void nopOperation(int position) {
		Operation op = operations.get(position);
		operations.set(position, new NopOperation(op.getTemplatePosition()));
	}

	public static MatchTemplateImpl createTemplateImpl(String matchLanguage, MatchTemplateResolver resolver, String templateName, BufferedReader br, Formatter[] formatters) throws IOException {
		if (matchLanguage.equals("plain")) {
			return new PlainMatchTemplateImpl(resolver, templateName, br, formatters);
		} else {
			throw new IllegalArgumentException("Unsupported match template language!");
		}
	}

	public String getCurrentTemplatePosition() {
		return currentTemplatePosition;
	}

	protected void setCurrentTemplatePosition(String currentTemplatePosition) {
		this.currentTemplatePosition = currentTemplatePosition;
	}

	public abstract String[] getSpecialTags();

	public Map<String, Label> getDefLabels() {
		return defLabels;
	}
}
