/*
 * Copyright (c) 2006 - 2011 Michael Schierl
 * All rights reserved.
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package net.sf.jmatchparser.template.engine.template;

import java.io.BufferedReader;
import java.io.IOException;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Set;
import java.util.regex.Pattern;

import net.sf.jmatchparser.template.Formatter;
import net.sf.jmatchparser.template.MatchTemplate.MatchTemplateResolver;
import net.sf.jmatchparser.template.engine.operation.CheckFinishedOperation;
import net.sf.jmatchparser.template.engine.operation.Operation.OperationResult;
import net.sf.jmatchparser.template.engine.template.command.plain.Command;

public class PlainMatchTemplateImpl extends MatchTemplateImpl {

	private final Set<String> usedNames = new HashSet<String>();
	private final Set<String> defines = new HashSet<String>();
	private final String[] specialTags = new String[] { "", "", null, null };

	public PlainMatchTemplateImpl(MatchTemplateResolver resolver, String templateName, BufferedReader br, Formatter[] formatters) throws IOException {
		super(formatters);
		load(resolver, "", templateName, br);
		appendOperation(new CheckFinishedOperation(getCurrentTemplatePosition(), OperationResult.FINISHED));
	}

	private void load(MatchTemplateResolver resolver, String prefix, String templateName, BufferedReader br) throws IOException {
		if (br == null)
			br = resolver.resolve(templateName);
		if (!includedTemplateFileNames.contains(templateName))
			includedTemplateFileNames.add(templateName);
		List<PlainBlockCommandState> openBlockStates = new ArrayList<PlainBlockCommandState>();
		// List<String> fileLines = new ArrayList<String>();
		String templateNameUnique = prefix + templateName;
		while (usedNames.contains(templateNameUnique)) {
			templateNameUnique += "_";
		}
		usedNames.add(templateNameUnique);
		String line;
		boolean active = true;
		List<Boolean> activeStack = new ArrayList<Boolean>();
		List<String> activeStackTemplatePositions = new ArrayList<String>();
		int lineNumber = 0;
		while ((line = br.readLine()) != null) {
			lineNumber++;
			setCurrentTemplatePosition("[" + templateNameUnique + ":" + lineNumber + "]\t" + line);
			try {
				if ((active || line.trim().startsWith("$")) && line.trim().length() > 0 && !line.trim().startsWith("#")) {
					while (line.startsWith(" ") || line.startsWith("\t"))
						line = line.substring(1);
					int pos = line.indexOf(" ");
					int pos2 = line.indexOf("\t");
					if (pos2 != -1 && (pos == -1 || pos > pos2))
						pos = pos2;
					String commandName, parameters;
					if (pos == -1) {
						commandName = line.toLowerCase();
						parameters = "";
					} else {
						commandName = line.substring(0, pos).toLowerCase();
						parameters = line.substring(pos + 1);
					}
					if (commandName.startsWith("$")) {
						if (commandName.equals("$include")) {
							if (active) {
								String tp = getCurrentTemplatePosition();
								load(resolver, templateNameUnique + "|", parameters, null);
								setCurrentTemplatePosition(tp);
							}
						} else if (commandName.equals("$ifdef")) {
							activeStackTemplatePositions.add(getCurrentTemplatePosition());
							activeStack.add(active);
							active &= defines.contains(parameters);
						} else if (commandName.equals("$ifndef")) {
							activeStackTemplatePositions.add(getCurrentTemplatePosition());
							activeStack.add(active);
							active &= !defines.contains(parameters);
						} else if (commandName.equals("$endif")) {
							activeStackTemplatePositions.remove(activeStackTemplatePositions.size() - 1);
							active = activeStack.remove(activeStack.size() - 1);
						} else if (commandName.equals("$define")) {
							if (active) {
								defines.add(parameters);
							}
						} else if (commandName.equals("$undef")) {
							if (active) {
								defines.remove(parameters);
							}
						} else if (commandName.equals("$specialtags")) {
							if (active) {
								String[] parts = parameters.substring(1).split(Pattern.quote(parameters.substring(0, 1)), -1);
								if (parts[parts.length - 1].length() > 0)
									throw new RuntimeException("Separator missing at end");
								if (parts.length == 3) {
									specialTags[0] = parts[0];
									specialTags[1] = parts[1];
									specialTags[2] = null;
									specialTags[3] = null;
								} else if (parts.length == 5) {
									specialTags[0] = parts[0];
									specialTags[1] = parts[1];
									specialTags[2] = parts[2];
									specialTags[3] = parts[3];
								} else {
									throw new RuntimeException("Invalid number of tags");
								}
							}
						} else {
							throw new RuntimeException("Unsupported preprocessor command: " + commandName);
						}
					} else {
						Command cmd = Command.getCommand(commandName);
						boolean parsed = false;
						if (cmd != null) {
							PlainBlockCommandState bcs = cmd.parse(this, parameters);
							if (bcs != null)
								openBlockStates.add(bcs);
							parsed = true;
						} else if (openBlockStates.size() > 0) {
							PlainBlockCommandState bcs = openBlockStates.remove(openBlockStates.size() - 1);
							if (bcs.canParse(commandName)) {
								bcs = bcs.parseCommand(this, commandName, parameters);
								parsed = true;
								if (bcs != null)
									openBlockStates.add(bcs);
							}
						}
						if (!parsed) {
							throw new RuntimeException("Unsupported command: " + commandName);
						}
					}
				}
			} catch (RuntimeException ex) {
				throw new RuntimeException("Cannot parse line:\n\t" + getCurrentTemplatePosition(), ex);
			}
		}
		br.close();
		if (openBlockStates.size() != 0) {
			String pos = openBlockStates.get(openBlockStates.size() - 1).getStartPosition();
			throw new IOException("Unclosed block at position:\n\t" + pos);
		}
		if (activeStackTemplatePositions.size() != 0) {
			String templatePosition = activeStackTemplatePositions.get(activeStackTemplatePositions.size() - 1);
			throw new IOException("Unclosed preprocessor command:\n\t" + templatePosition);
		}
		setCurrentTemplatePosition("[" + templateNameUnique + ":" + (lineNumber + 1) + "]\t[End of file]");
	}

	@Override
	public String[] getSpecialTags() {
		return specialTags;
	}
}
